import os
import sys
import asyncio
import traceback
import aiohttp
from datetime import datetime

MCP_HOST = os.getenv("MCP_HOST", "http://localhost:7860")
BASE_URL = f"{MCP_HOST}"
BACKEND_URL_SSE = f"{BASE_URL}/api/v1/mcp/sse"
BACKEND_URL_MSG = f"{BASE_URL}/api/v1/mcp/"
DEBUG = os.getenv("DEBUG", "false").lower() == "true"

# Add message queue at the top with other globals
message_queue = []
message_endpoint = None

# Update the connection logic to derive SSL from URL scheme
use_ssl = MCP_HOST.lower().startswith("https://")

# Create SSL context or None based on URL scheme
ssl_context = False if not use_ssl else True

if sys.platform == "win32":
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

def debug(message):
    """Output debug messages to stderr with timestamps."""
    if DEBUG:
        timestamp = datetime.now().strftime('%Y-%m-%d %H:%M:%S.%f')[:-3]
        print(f"[{timestamp}] {message}", file=sys.stderr)

async def connect_sse_backend():
    """Establish persistent SSE connection to MCP server."""
    global message_endpoint
    retry_count = 0
    max_retries = 3
    
    try:
        # Create session with explicit SSL context and verify settings are applied
        connector = aiohttp.TCPConnector(ssl=ssl_context, force_close=True)
        debug(f"SSE: Creating connector with ssl={ssl_context}, connector={connector}")
        
        async with aiohttp.ClientSession(connector=connector, trust_env=False) as session:
            while retry_count < max_retries:
                try:
                    debug(f"SSE: Attempt {retry_count + 1} connecting to {BACKEND_URL_SSE}")
                    debug(f"SSE: Using connector settings: ssl={ssl_context}, host={MCP_HOST}")
                    
                    async with session.get(BACKEND_URL_SSE, ssl=ssl_context) as response:
                        if response.status != 200:
                            raise Exception(f"SSE: Connection failed with status {response.status}")

                        debug("SSE: Connected successfully")

                        # Read and process SSE messages
                        async for line in response.content:
                            if line:
                                message = line.decode().strip()
                                debug(f"SSE <<< {message}")
                                
                                if message.startswith("event: endpoint"):
                                    continue
                                elif message.startswith("data: ") and message_endpoint is None:
                                    endpoint = message[6:]
                                    message_endpoint = f"{BASE_URL}{endpoint}"
                                    debug(f"SSE: Message endpoint set to: {message_endpoint}")
                                    
                                    # Process any queued messages
                                    if message_queue:
                                        debug(f"QUEUE: Processing {len(message_queue)} queued messages")
                                        for queued_message in message_queue:
                                            await process_message(session, queued_message)
                                        message_queue.clear()
                                elif message.startswith("data: "):
                                    response_data = message[6:]
                                    debug(f"SSE >>> {response_data}")
                                    print(response_data, flush=True)

                except aiohttp.ClientConnectorError as e:
                    debug(f"Connection failed: {e}")
                    debug(f"Check network settings for {BACKEND_URL_SSE}")
                    retry_count += 1
                    if retry_count < max_retries:
                        await asyncio.sleep(1)
                    else:
                        raise
                except aiohttp.ClientSSLError as ssl_err:
                    debug(f"SSL Error: {ssl_err}")
                    debug("Ensure SSL certificate and URL scheme match.")
                    raise
                except Exception as e:
                    debug(f"Unexpected error: {e}")
                    retry_count += 1
                    if retry_count < max_retries:
                        await asyncio.sleep(1)
                    else:
                        raise
    except Exception as e:
        debug(f"--- SSE backend disc./error: {str(e)}")
        debug(f"Full connection details: MCP_HOST={MCP_HOST}, ssl={ssl_context}")
        raise


async def process_message(session, message):
    """Forward received message to the MCP server."""
    if not message_endpoint:
        debug(f"QUEUE: Message queued (no endpoint): {message}")
        message_queue.append(message)
        return
        
    debug(f"MSG >>> {message.strip()}")
    try:
        async with session.post(message_endpoint, data=message, headers={"Content-Type": "application/json"}) as resp:
            debug(f"MSG <<< Status: {resp.status}")
            if resp.status != 202:
                debug(f"MSG <<< Error: Unexpected status {resp.status}")
    except Exception as e:
        debug(f"MSG <<< Error: {e}")
        debug(f"Full exception: {traceback.format_exc()}")

async def run_bridge():
    """Run the bridge."""
    try:
        # Start the SSE connection in a background task
        asyncio.create_task(connect_sse_backend())

        # Use same SSL settings for message processing session
        connector = aiohttp.TCPConnector(ssl=ssl_context)
        async with aiohttp.ClientSession(connector=connector) as session:
            debug("-- MCP stdio to SSE gw running")

            # Read stdin synchronously using a ThreadPoolExecutor
            loop = asyncio.get_running_loop()

            def read_stdin_sync():
                return sys.stdin.read()

            while True:
                # Read a line synchronously from stdin
                message = await loop.run_in_executor(None, sys.stdin.readline)
                if not message:  # End of input
                    break
                await process_message(session, message.strip())
    except Exception as error:
        debug(f"Fatal error running server: {error}")
        trace = traceback.format_exc()
        debug(f"Traceback: {trace}")
        sys.exit(1)

def app():
    asyncio.run(run_bridge())

if __name__ == "__main__":
    asyncio.run(app())

